"""The april_vision CLI."""
import argparse
import importlib
import logging

from april_vision._version import version

subcommands = [
    "tools",
    "annotate_image",
    "annotate_video",
    "calibrate",
    # "camera_benchmark",
    "live",
    # "marker_benchmark",
    "marker_generator",
    "vision_debug",
]


def build_argparser() -> argparse.ArgumentParser:
    """Load subparsers from available subcommands."""
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(required=True)
    for command in subcommands:
        mod_name = f"{__package__}.{command}"
        importlib.import_module(mod_name).create_subparser(subparsers)

    parser.add_argument(
        '--version', action='version', version=version, help="Print package version")
    parser.add_argument('--debug', action='store_true', help="Enable debug logging")

    return parser


def setup_logger(debug: bool = False) -> None:
    """Output all loggers to console with custom format at level INFO or DEBUG."""
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(formatter)

    # log from all loggers to stdout
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)
    root_logger.addHandler(console_handler)

    if debug:
        root_logger.setLevel(logging.DEBUG)


def main() -> None:
    """CLI entry-point."""
    parser = build_argparser()
    args = parser.parse_args()
    setup_logger(debug=args.debug)

    if "func" in args:
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



"""
A charuco calibration script.

From: https://gist.github.com/naoki-mizuno/d25cbc3c59228291cabe50529d70894c
"""
import argparse
import logging
import sys
from datetime import datetime
from typing import Any, Iterable, List, Tuple

import cv2
from numpy.typing import NDArray

LOGGER = logging.getLogger(__name__)


def read_chessboards(
    frames: Iterable[NDArray],
    aruco_dict: Any,
    board: Any,
) -> Tuple[Any, Any, Any]:
    """Charuco base pose estimation."""
    all_corners = []
    all_ids = []

    for frame in frames:
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        corners, ids, rejectedImgPoints = cv2.aruco.detectMarkers(gray, aruco_dict)

        if len(corners) > 0:
            ret, c_corners, c_ids = cv2.aruco.interpolateCornersCharuco(
                corners, ids, gray, board)
            # ret is the number of detected corners
            if ret > 0:
                all_corners.append(c_corners)
                all_ids.append(c_ids)
        else:
            LOGGER.error("Failed!")

    imsize = gray.shape
    return all_corners, all_ids, imsize


def capture_camera(
    cap: cv2.VideoCapture,
    num: int = 1,
    mirror: bool = False,
    size: Tuple[int, int] = None,
) -> List[NDArray]:
    """Capture frames to be used for calibration."""
    frames = []
    LOGGER.info("Press space to capture frame.")

    while True:
        ret, frame = cap.read()

        if mirror is True:
            frame = cv2.flip(frame, 1)

        if size is not None and len(size) == 2:
            frame = cv2.resize(frame, size)

        cv2.imshow("Frame", frame)

        k = cv2.waitKey(1)
        if k == 27:  # Esc
            break
        elif k == 10 or k == 32:  # Enter or Space
            frames.append(frame)
            LOGGER.info("Frame captured!")
            if len(frames) == num:
                break

    return frames


def main(args: argparse.Namespace) -> None:
    """Charuco calibration."""
    try:
        import cv2.aruco  # type: ignore
    except ImportError:
        LOGGER.critical("Calibration requires the opencv-contrib-python package")
        sys.exit(1)

    aruco_dict = cv2.aruco.Dictionary_get(cv2.aruco.DICT_4X4_50)
    board = cv2.aruco.CharucoBoard_create(8, 6, 0.03, 0.023, aruco_dict)

    dev_num = args.camera
    video_dev = cv2.VideoCapture(dev_num)

    if args.resolution:
        video_dev.set(cv2.CAP_PROP_FRAME_WIDTH, args.resolution[0])
        video_dev.set(cv2.CAP_PROP_FRAME_HEIGHT, args.resolution[1])

    frames = capture_camera(video_dev, args.frame_count)
    if len(frames) == 0:
        LOGGER.error("No frame captured")
        sys.exit(1)
    all_corners, all_ids, imsize = read_chessboards(frames, aruco_dict, board)
    ret, camera_matrix, dist_coeff, rvec, tvec = cv2.aruco.calibrateCameraCharuco(
        all_corners, all_ids, board, imsize, None, None,
    )

    LOGGER.info(f"> Camera matrix\n{camera_matrix}")
    LOGGER.info(f"> Distortion coefficients\n{dist_coeff}")

    width = int(video_dev.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(video_dev.get(cv2.CAP_PROP_FRAME_HEIGHT))
    LOGGER.info(f"> Resolution\n{width}x{height}")

    video_dev.release()

    LOGGER.info("Generating calibration XML file")
    output_filename = args.cal_filename
    if not args.cal_filename.lower().endswith(".xml"):
        output_filename += ".xml"

    file = cv2.FileStorage(args.cal_filename, cv2.FILE_STORAGE_WRITE)

    calibrationDate = datetime.now().strftime("%a %d %b %Y %H:%M:%S")
    file.write("calibrationDate", calibrationDate)

    file.write("framesCount", args.frame_count)

    file.startWriteStruct("cameraResolution", cv2.FILE_NODE_SEQ)
    file.write("", width)
    file.write("", height)
    file.endWriteStruct()

    file.write("cameraMatrix", camera_matrix)
    file.write("dist_coeffs", dist_coeff)

    if args.vidpid is not None:
        # Wrap the str in quotes so it is outputed and loaded correctly
        file.write("vidpid", f'"{args.vidpid}"')

    file.release()


def create_subparser(subparsers: argparse._SubParsersAction) -> None:
    """Calibrate command parser."""
    parser = subparsers.add_parser(
        "calibrate",
        help="Generate camera calibration",
        description="""
        Generate camera calibration using a ChArUco board.

        Generate a board with:
        https://calib.io/pages/camera-calibration-pattern-generator
        260x200mm, 6x8 squares, 30mm checkers.
        """)
    parser.add_argument("camera", type=int, help="The camera index")
    parser.add_argument(
        "--resolution", type=int, nargs=2, default=None,
        metavar=("WIDTH", "HEIGHT"),
        help="Force camera resolution")
    parser.add_argument(
        "-n", "--frame_count", type=int, default=15,
        help="Number of frames to use for calbration")
    parser.add_argument(
        "--vidpid", type=str, default=None,
        help="VID and PID to be put on the calibration file")
    parser.add_argument("cal_filename", type=str, help="Filename of output calibration file")

    parser.set_defaults(func=main)



"""
Open cameras and measures the performance.

Measures the FPS for different resolutions.
"""
import argparse

# TODO Impliment this


def main(args: argparse.Namespace) -> None:
    """Measure the FPS for different resolutions."""
    raise NotImplementedError("Camera benchmark is not implemented.")


def create_subparser(subparsers: argparse._SubParsersAction) -> None:
    """Camera_benchmark command parser."""
    parser = subparsers.add_parser("camera_benchmark")

    parser.set_defaults(func=main)



"""An AprilTags wrapper with camera discovery and axis conversion."""
import os

os.environ["OPENCV_VIDEOIO_MSMF_ENABLE_HW_TRANSFORMS"] = "0"

from ._version import __version__
from .calibrations import calibrations
from .detect_cameras import CalibratedCamera, find_cameras
from .frame_sources import FrameSource, USBCamera
from .helpers import generate_marker_size_mapping
from .marker import (CartesianCoordinates, Marker, Orientation,
                     PixelCoordinates, SphericalCoordinate)
from .utils import Frame
from .vision import Processor

__all__ = [
    '__version__',
    'CalibratedCamera',
    'CartesianCoordinates',
    'Frame',
    'FrameSource',
    'Marker',
    'Orientation',
    'PixelCoordinates',
    'Processor',
    'SphericalCoordinate',
    'USBCamera',
    'calibrations',
    'find_cameras',
    'generate_marker_size_mapping',
]



import base64
from threading import Thread
from typing import Callable, List, Optional

import cv2
import numpy as np
from numpy.typing import NDArray

from ..marker import Marker
from ..utils import Frame
from ..vision import Processor


class Base64Sender:
    """
    A helper class to encode and send frame data using the publish callback.

    The image data is a base64 encoded JPEG bytestring.
    :param annotated: Controls whether images are annotated with detections
                      before sending. This will copy the frame prior to annotation.
    :param threaded: Controls whether encoding and sending the image is processed
                     in a thread or blocks.
    """
    def __init__(
        self,
        publish_callback: Callable[[str, bytes], None],
        *,
        annotated: bool = True,
        threaded: bool = True,
        aruco_orientation: bool = True,
    ):
        self._publish_callback = publish_callback
        self.use_threads = threaded
        self.send_annotated = annotated
        self._processor = Processor(aruco_orientation=aruco_orientation)

    def annotated_frame_hook(self, frame: Frame, markers: List[Marker]) -> None:
        """
        The hook function called in the detection hook.

        Optionally annotates a copy of the frame, possibly in a background
        thread, depending on the instance settings.
        """
        copied_frame = Frame(
            np.array(frame.grey_frame, copy=True),
            np.array(frame.colour_frame, copy=True),
        )
        if self.send_annotated:
            output_frame = self._processor._annotate(copied_frame, markers)
        else:
            output_frame = copied_frame

        if self.use_threads:
            thread = Thread(
                name="Image send",
                target=self.encode_and_send,
                args=(output_frame.colour_frame,),
                daemon=True,
            )
            thread.start()
        else:
            self.encode_and_send(output_frame.colour_frame)

    def encode_and_send(self, frame: NDArray[np.uint8]) -> None:
        """
        Handle converting a frame to a base64 bytestring and sending it using
        the publish callback.

        Can be run as a thread target.
        """
        if self._publish_callback is None:
            return

        encoded_frame = self.base64_encode_frame(frame)
        if encoded_frame is None:
            return

        encoded_frame = b'data:image/jpeg;base64, ' + encoded_frame

        # The message is sent in a background thread
        try:
            self._publish_callback(
                'camera/annotated' if self.send_annotated else 'camera/image',
                encoded_frame,
            )
        except ValueError:
            pass

    def base64_encode_frame(self, frame: NDArray[np.uint8]) -> Optional[bytes]:
        """Convert image frame to base64 bytestring."""
        ret, image = cv2.imencode('.jpg', frame)
        if not ret:
            return None
        image_bytes = image.tobytes()
        return base64.b64encode(image_bytes)



"""
Live camera demonstration.

Opens the camera and does live marker detection.
Can add overlays of fps, marker annotation, marker distance, etc..
Option to save the current view of the camera.
"""
import argparse
import logging
import os
from math import degrees
from time import perf_counter

import cv2

from ..calibrations import calibrations
from ..detect_cameras import find_cameras
from ..frame_sources import USBCamera
from ..marker import MarkerType
from ..utils import RollingAverage, annotate_text, normalise_marker_text
from ..vision import Processor

LOGGER = logging.getLogger(__name__)


def main(args: argparse.Namespace) -> None:
    """Live camera demonstration."""
    avg_fps = RollingAverage(50)
    prev_frame_time: float = 0
    file_num = 1
    if args.id is None:
        cameras = find_cameras(calibrations, include_uncalibrated=True)
        try:
            camera = cameras[0]
        except IndexError:
            LOGGER.fatal("No cameras found")
            return
        source = USBCamera.from_calibration_file(
            camera.index, camera.calibration, camera.vidpid)
    else:
        source = USBCamera(args.id, (1280, 720))
    cam = Processor(
        source,
        tag_family=args.tag_family,
        quad_decimate=args.quad_decimate,
        tag_sizes=float(args.tag_size) / 1000,
        calibration=source.calibration,
    )

    LOGGER.info("Press S to save image, press Q to exit")

    while True:
        frame = cam._capture()
        markers = cam._detect(frame)

        if args.annotate:
            cam._annotate(frame, markers)

        new_frame_time = perf_counter()
        fps = 1 / (new_frame_time - prev_frame_time)
        prev_frame_time = new_frame_time
        avg_fps.new_data(fps)
        if args.fps:
            frame = annotate_text(
                frame, f"{avg_fps.average():.0f}", (7, 70),
                text_scale=3, text_colour=(100, 255, 0))

        if args.distance:
            for marker in markers:
                if marker.has_pose():
                    text_scale = normalise_marker_text(marker)

                    loc = (
                        int(marker.pixel_centre[0] - 80 * text_scale),
                        int(marker.pixel_centre[1] + 40 * text_scale),
                    )
                    frame = annotate_text(
                        frame, f"dist={marker.distance}mm", loc,
                        text_scale=0.8 * text_scale,
                        text_colour=(255, 191, 0),  # deep sky blue
                    )

        for marker in markers:
            output = "#{}, {}, ({}, {})".format(
                marker.id,
                marker.marker_type.value.strip('tag'),
                int(marker.pixel_centre.x),
                int(marker.pixel_centre.y),
            )

            if marker.has_pose():
                output += (
                    ", | b={} | r={}, θ={}, φ={} | "
                    "x={}, y={}, z={} | r={}, p={}, y={}"
                ).format(
                    int(marker.bearing),
                    int(marker.spherical.r),
                    int(degrees(marker.spherical.theta)),
                    int(degrees(marker.spherical.phi)),
                    int(marker.cartesian.x),
                    int(marker.cartesian.y),
                    int(marker.cartesian.z),
                    int(degrees(marker.orientation.roll)),
                    int(degrees(marker.orientation.pitch)),
                    int(degrees(marker.orientation.yaw)),
                )
            else:
                output += ", no values for pose estimation"

            LOGGER.info(output)

        cv2.imshow('image', frame.colour_frame)

        button = cv2.waitKey(1) & 0xFF
        if (button == ord('q')) or (button == 27):
            # Quit on q or ESC key
            break
        elif button == ord('s'):
            filename = f'saved_image{file_num:03d}.jpg'
            while os.path.exists(filename):
                file_num += 1
                filename = f'saved_image{file_num:03d}.jpg'

            cv2.imwrite(filename, frame.colour_frame)
            file_num += 1


def create_subparser(subparsers: argparse._SubParsersAction) -> None:
    """Live command parser."""
    parser = subparsers.add_parser(
        "live",
        description="Live camera demonstration with marker annotation.",
        help="Live camera demonstration with marker annotation.",
    )

    parser.add_argument(
        "--id", type=int, default=None, help="Override the camera index to use.")
    parser.add_argument(
        "--no_annotate", action='store_false', dest='annotate',
        help="Turn off marker annotation for detected markers.")
    parser.add_argument(
        '--fps', action='store_true',
        help="Display the frames per second that the preview is running at.")

    parser.add_argument(
        '--tag_family', default=MarkerType.APRILTAG_36H11.value,
        choices=[marker.value for marker in MarkerType],
        help="Set the marker family to detect, defaults to 'tag36h11'")
    parser.add_argument(
        '--quad_decimate', type=float, default=2,
        help="Set the level of decimation used in the detection stage")

    parser.add_argument(
        '--tag_size', type=int, default=0, help="The size of markers in millimeters")
    parser.add_argument(
        '--distance', action='store_true',
        help="Annotate frames with the distance to the marker.")

    parser.set_defaults(func=main)



import argparse

from tabulate import tabulate

from april_vision.calibrations import calibrations
from april_vision.detect_cameras import find_cameras


def main(args: argparse.Namespace) -> None:
    cameras = find_cameras(calibrations, include_uncalibrated=True)
    print(tabulate(cameras, headers="keys"))


def create_subparser(subparsers: argparse._SubParsersAction) -> None:
    """List out the available cameras connected to the system."""
    parser = subparsers.add_parser(
        "list_cameras",
        description=(
            "List out the available cameras connected to the system, "
            "this will also search your current directory for calibration files"
        ),
        help=(
            "List out the available cameras connected to the system, "
            "this will also search your current directory for calibration files"
        ),
    )

    parser.set_defaults(func=main)



from typing import List, NamedTuple, Tuple

import pyapriltags


class ApriltagFamily(NamedTuple):
    ncodes: int
    codes: List[int]
    width_at_border: int
    total_width: int
    reversed_border: bool
    nbits: int
    bits: List[Tuple[int, int]]
    h: int
    name: str

    def __str__(self) -> str:
        data = "Marker family info:\n"
        data += f"    Tag family:       {self.name}\n"
        data += f"    Valid markers:    0-{self.ncodes - 1}\n"
        data += f"    Total width:      {self.total_width}\n"
        data += f"    Width at border:  {self.width_at_border}\n"
        data += f"    Reversed border:  {self.reversed_border}\n"
        data += f"    Num bits:         {self.nbits}\n"
        data += f"    Hamming distance: {self.h}\n"
        return data


def get_tag_family(family: str) -> ApriltagFamily:
    """
    Use the C-types in pyapriltags to get the tag family object.
    This object contains the required data to draw all the tags for a given family.
    """
    d = pyapriltags.Detector(families=family)
    raw_tag_data = d.tag_families[family].contents

    tag_data = ApriltagFamily(
        ncodes=raw_tag_data.ncodes,
        codes=[raw_tag_data.codes[i] for i in range(raw_tag_data.ncodes)],
        width_at_border=raw_tag_data.width_at_border,
        total_width=raw_tag_data.total_width,
        reversed_border=raw_tag_data.reversed_border,
        nbits=raw_tag_data.nbits,
        bits=[
            (raw_tag_data.bit_x[i], raw_tag_data.bit_y[i])
            for i in range(raw_tag_data.nbits)
        ],
        h=raw_tag_data.h,
        name=raw_tag_data.name.decode("utf-8"),
    )
    return tag_data


def parse_ranges(ranges: str) -> List[int]:
    """
    Parse a comma seprated list of numbers which may include ranges
    specified as hyphen-separated numbers.
    From https://stackoverflow.com/questions/6405208
    """
    result: List[int] = []
    for part in ranges.split(","):
        if "-" in part:
            a_, b_ = part.split("-")
            a, b = int(a_), int(b_)
            result.extend(range(a, b + 1))
        else:
            a = int(part)
            result.append(a)
    return result



import argparse
import importlib

subcommands = [
    "family_details",
    "list_cameras",
]


def create_subparser(subparsers: argparse._SubParsersAction) -> None:
    parser = subparsers.add_parser(
        "tools",
        description="A collection of useful tools",
        help="A collection of useful tools",
    )

    subparsers = parser.add_subparsers(required=True)
    for command in subcommands:
        mod_name = f"{__package__}.{command}"
        importlib.import_module(mod_name).create_subparser(subparsers)



"""
Generate marker PDFs.

Takes a set of parameters and generates marker PDFs.
"""
import argparse
import logging
from enum import Enum
from pathlib import Path
from typing import Tuple

import numpy as np
from numpy.typing import NDArray
from PIL import Image, ImageDraw, ImageFont, ImageOps
from pypdf import PdfMerger

from ..marker import MarkerType
from .utils import ApriltagFamily, get_tag_family, parse_ranges

LOGGER = logging.getLogger(__name__)

DPI = 72
BORDER_FILL = "lightgrey"


def mm_to_pixels(mm: int) -> int:
    """
    Convert millimeters to pixels
    """
    inches = mm / 25.4
    return int(inches * DPI)


class PageSize(Enum):
    A3 = (297, 420)
    A4 = (210, 297)

    @property
    def pixels(self) -> Tuple[int, int]:
        return (
            mm_to_pixels(self.value[0]),
            mm_to_pixels(self.value[1]),
        )


class PageMode(Enum):
    SINGLE = 1
    TILE = 2
    CROP = 3
    SPLIT = 4


def generate_tag_array(tag_data: ApriltagFamily, tag_id: int) -> NDArray:
    """
    Uses the tag family object to generate a marker, returns this data as a 2d numpy array
    where each of the cells is 1 pixel of the marker.
    """
    # Create grid of tag size
    dim = tag_data.total_width
    tag = np.ones((dim, dim), dtype=np.uint8) * 255

    # Find position of origin in grid
    border_edge = int((tag_data.total_width / 2.0) - (tag_data.width_at_border / 2.0))

    # Draw black/white boarder lines
    if tag_data.reversed_border:
        left_border_edge = border_edge - 1
        right_border_edge = border_edge + tag_data.width_at_border
    else:
        left_border_edge = border_edge
        right_border_edge = border_edge + tag_data.width_at_border - 1

    for i in range(left_border_edge, right_border_edge+1):
        tag[left_border_edge][i] = 0
        tag[i][left_border_edge] = 0
        tag[right_border_edge][i] = 0
        tag[i][right_border_edge] = 0

    # Fill in pixels, black if bit index is zero
    max_index = tag_data.nbits - 1
    for i, (x, y) in enumerate(tag_data.bits):
        binary = bool(tag_data.codes[tag_id] & (1 << (max_index - i)))
        tag[y+border_edge][x+border_edge] = 255 if binary else 0

    return tag


def generate_tag_tile(
    tag_data: ApriltagFamily,
    args: argparse.Namespace,
    tag_id: int,
) -> Image.Image:
    """
    Uses the tag family object to generate a marker image.
    The marker is scaled to the correct size and is annotated with a border and text.
    """
    # Calculate the overall marker size
    pixel_size = args.tag_size // tag_data.width_at_border
    required_size = pixel_size * tag_data.total_width

    # Generate marker image and resize
    tag_array = generate_tag_array(tag_data, tag_id)

    if args.aruco_orientation:
        # Rotate by 180deg to match the aruco format
        tag_array = np.rot90(tag_array, k=2)

    marker_image = Image.fromarray(tag_array)
    resized_image = marker_image.resize(
        (mm_to_pixels(required_size), mm_to_pixels(required_size)),
        resample=0
    )

    # Add grey border line
    bordered_image = ImageOps.expand(
        resized_image,
        border=args.border_width,
        fill=BORDER_FILL
    )
    img_size = bordered_image.size[0]
    image_draw = ImageDraw.Draw(bordered_image)

    # Add center tick marks
    line_start = (img_size // 2) - (args.border_width // 2)

    # Top
    image_draw.line(
        [line_start, 0, line_start, args.tick_length],
        width=args.border_width,
        fill=BORDER_FILL,
    )

    # Left
    image_draw.line(
        [0, line_start, args.tick_length, line_start],
        width=args.border_width,
        fill=BORDER_FILL,
    )

    # Bottom
    image_draw.line(
        [line_start, img_size - args.tick_length, line_start, img_size],
        width=args.border_width,
        fill=BORDER_FILL,
    )

    # Right
    image_draw.line(
        [img_size - args.tick_length, line_start, img_size, line_start],
        width=args.border_width,
        fill=BORDER_FILL,
    )

    # Add text to the image
    marker_sqaure_size = mm_to_pixels(pixel_size)

    # Draw tag ID number in corner of white boarder
    border_edge = int((tag_data.total_width / 2.0) - (tag_data.width_at_border / 2.0))
    if tag_data.reversed_border:
        id_pos = (marker_sqaure_size * border_edge) + (marker_sqaure_size // 2)
    else:
        id_pos = (marker_sqaure_size * (border_edge - 1)) + (marker_sqaure_size // 2)

    if args.no_number is False:
        image_draw.text(
            (id_pos, id_pos),
            f"{tag_id}",
            fill=BORDER_FILL,
            anchor="mm",
            font=ImageFont.truetype(args.font, args.number_size),
        )

    # Expand the tile to add a white border with a width of 1 marker square
    text_border_image = ImageOps.expand(
        bordered_image,
        border=marker_sqaure_size,
        fill="white"
    )
    image_draw_expand = ImageDraw.Draw(text_border_image)

    # Draw text outside the marker
    image_draw_expand.text(
        (marker_sqaure_size, text_border_image.size[0] - (marker_sqaure_size // 2)),
        args.description_format.format(
            marker_type=tag_data.name, marker_id=tag_id
        ),
        anchor="lm",
        font=ImageFont.truetype(args.font, args.number_size),
    )

    return text_border_image


def main(args: argparse.Namespace) -> None:
    """
    Create markers and marker PDFs using the parameters provided on the command line.
    """
    tag_data = get_tag_family(args.tag_family)
    LOGGER.info(tag_data)

    # Get list of markers we want to make
    if args.range == "ALL":
        marker_ids = [num for num in range(tag_data.ncodes)]
    else:
        try:
            marker_ids = parse_ranges(args.range)
        except ValueError:
            LOGGER.error("Invalid marker number range provided")
            exit(1)

    if (max(marker_ids) > (tag_data.ncodes - 1)) or (min(marker_ids) < 0):
        LOGGER.error("Supplied marker number lies outside permitted values for marker family")
        LOGGER.error(f"Permitted marker range: 0-{tag_data.ncodes - 1}")
        exit(1)

    LOGGER.info(f"Generating {len(marker_ids)} markers")

    # Output page size and location
    page_size = PageSize[args.page_size]

    output_dir = args.output_dir.resolve()
    output_dir.mkdir(exist_ok=True, parents=True)

    generated_files = []

    # Generate markers
    for marker_id in marker_ids:
        LOGGER.info(f"Creating marker: {marker_id}")

        image_tile = generate_tag_tile(tag_data, args, tag_id=marker_id)
        tile_size = image_tile.size[0]

        page_mode = PageMode[args.page_mode]

        if page_mode == PageMode.SINGLE:
            output_img = Image.new("RGB", page_size.pixels, (255, 255, 255))

            if args.bottom_padding is None:
                x_loc = (page_size.pixels[0] - tile_size) // 2
                y_loc = (page_size.pixels[1] - tile_size) // 2
                output_img.paste(image_tile, (x_loc, y_loc))
            else:
                x_loc = (page_size.pixels[0] - tile_size) // 2

                # Awkward math as we extended the tile beyond the border as extra area for
                # the text, The size of the image tile does not match the border
                marker_sqaure_size = mm_to_pixels(args.tag_size // tag_data.width_at_border)
                y_loc = (page_size.pixels[1]
                         - mm_to_pixels(args.bottom_padding)
                         - (marker_sqaure_size * (tag_data.total_width + 1))
                         - args.border_width
                         )
                output_img.paste(image_tile, (x_loc, y_loc))

            generated_files.append(output_dir / args.filename.format(id=marker_id))
            try:
                output_img.save(
                    output_dir / args.filename.format(id=marker_id),
                    quality=100,
                    dpi=(DPI, DPI),
                )
            except ValueError as error:
                LOGGER.error(f"Invalid output file format: '{args.filename}'")
                LOGGER.error(error)
                exit(1)

        elif page_mode == PageMode.TILE:
            output_img = Image.new("RGB", page_size.pixels, (255, 255, 255))

            for row in range(args.row_num):
                for col in range(args.column_num):
                    col_space_px = mm_to_pixels(args.column_spacing)
                    row_space_px = mm_to_pixels(args.row_spacing)
                    x_loc = ((col_space_px + tile_size) * col) + col_space_px
                    y_loc = ((row_space_px + tile_size) * row) + row_space_px

                    output_img.paste(image_tile, (x_loc, y_loc))

            generated_files.append(output_dir / args.filename.format(id=marker_id))
            try:
                output_img.save(
                    output_dir / args.filename.format(id=marker_id),
                    quality=100,
                    dpi=(DPI, DPI),
                )
            except ValueError as error:
                LOGGER.error(f"Invalid output file format: '{args.filename}'")
                LOGGER.error(error)
                exit(1)

        elif page_mode == PageMode.CROP:
            if args.crop_size is None:
                pixel_size = args.tag_size // tag_data.width_at_border
                required_size = pixel_size * tag_data.total_width
                crop_pixels = mm_to_pixels(required_size) + args.border_width
            else:
                crop_pixels = mm_to_pixels(args.crop_size)

            output_img = Image.new("RGB", (crop_pixels, crop_pixels), (255, 255, 255))
            x_loc = (crop_pixels - tile_size) // 2
            y_loc = (crop_pixels - tile_size) // 2
            output_img.paste(image_tile, (x_loc, y_loc))

            try:
                output_img.save(
                    output_dir / args.filename.format(id=marker_id),
                    quality=100,
                    dpi=(DPI, DPI),
                )
            except ValueError as error:
                LOGGER.error(f"Invalid output file format: '{args.filename}'")
                LOGGER.error(error)
                exit(1)

        elif page_mode == PageMode.SPLIT:
            output_img_1 = Image.new("RGB", page_size.pixels, (255, 255, 255))
            output_img_2 = Image.new("RGB", page_size.pixels, (255, 255, 255))

            # Draw text outside the marker in top right corner
            marker_sqaure_size = mm_to_pixels(args.tag_size // tag_data.width_at_border)

            image_draw_text = ImageDraw.Draw(image_tile)
            image_draw_text.text(
                (tile_size - marker_sqaure_size, (marker_sqaure_size // 2)),
                args.description_format.format(
                    marker_type=tag_data.name, marker_id=marker_id
                ),
                anchor="rm",
                font=ImageFont.truetype(args.font, args.number_size),
            )

            image_half_left = image_tile.crop((0, 0, tile_size // 2, tile_size))
            image_half_right = image_tile.crop((tile_size // 2, 0, tile_size, tile_size))

            x_loc = (page_size.pixels[0] - image_half_left.size[0]) // 2
            y_loc = (page_size.pixels[1] - image_half_left.size[1]) // 2

            output_img_1.paste(image_half_left, (x_loc, y_loc))
            output_img_2.paste(image_half_right, (x_loc, y_loc))

            generated_files.append(output_dir / args.filename.format(id=marker_id))
            try:
                output_img_1.save(
                    output_dir / args.filename.format(id=marker_id),
                    quality=100,
                    dpi=(DPI, DPI),
                    save_all=True,
                    append_images=[
                        output_img_2
                    ],
                )
            except ValueError as error:
                LOGGER.error(f"Invalid output file format: '{args.filename}'")
                LOGGER.error(error)
                exit(1)

    if args.merge_pdf is not None:
        if args.filename.lower().endswith(".pdf"):
            LOGGER.info("Starting to merge PDFs")
            merger = PdfMerger()
            for pdf in generated_files:
                merger.append(pdf)

            if args.merge_pdf.lower().endswith(".pdf"):
                merger.write(output_dir / args.merge_pdf)
            else:
                merger.write(output_dir / (args.merge_pdf + '.pdf'))

            merger.close()
            LOGGER.info("Merge PDF complete")
        else:
            LOGGER.error((
                "PDF merge was enabled but no PDFs were generated, "
                f"output format '{args.filename}'"
            ))


def create_subparser(subparsers: argparse._SubParsersAction) -> None:
    """Marker_generator command parser."""

    parser = subparsers.add_parser(
        "marker_generator",
        description="Generate a PDF containing markers",
        help="Generate a PDF containing markers",
    )

    # Args for output filename/location
    parser.add_argument(
        "output_dir",
        help="The directory to save the output files to",
        type=Path,
    )
    parser.add_argument(
        "--filename",
        type=str,
        help=(
            "Output filename. `id` available for string format replacement "
            "(default: %(default)s)"
        ),
        default="{id}.pdf",
    )
    parser.add_argument(
        "--merge_pdf",
        help="Merge all the outputted PDFs into a single PDF of the provided filename",
        type=str,
        default=None,
    )

    # Args for modifying type and size
    tag_group = parser.add_argument_group('TAGS')
    tag_group.add_argument(
        "--tag_family", default=MarkerType.APRILTAG_36H11.value,
        choices=[marker.value for marker in MarkerType],
        help="Set the marker family to detect, defaults to 'tag36h11'",
    )
    tag_group.add_argument(
        "--tag_size",
        help="The size of markers in millimeters (default: %(default)s)",
        default=100,
        type=int,
    )
    tag_group.add_argument(
        "--no_aruco_orientation",
        help="Rotate marker 180 for standard orientation",
        action="store_false",
        dest="aruco_orientation",
    )
    tag_group.add_argument(
        "--range",
        help="Marker ids to output, can use '-' or ',' to specify lists and ranges",
        default="ALL",
    )

    # Args for modifying page size and marker layout
    page_group = parser.add_argument_group('PAGE')
    page_group.add_argument(
        "--page_size",
        type=str,
        help="Page size. (default: %(default)s)",
        choices=sorted([size.name for size in PageSize]),
        default="A4",
    )
    page_group.add_argument(
        "--page_mode",
        type=str,
        help="Page generation method. (default: %(default)s)",
        choices=sorted([mode.name for mode in PageMode]),
        default="SINGLE",
    )

    # Options for a SINGLE layout
    single_group = parser.add_argument_group('mode: SINGLE')
    single_group.add_argument(
        "--bottom_padding",
        help="Distance in mm between bottom border of marker and the bottom of the page",
        default=None,
        type=int,
    )

    # Options for a TILE layout
    tile_group = parser.add_argument_group('mode: TILE')
    tile_group.add_argument(
        "--column_num",
        help="Set number of columns of markers",
        default=1,
        type=int,
    )
    tile_group.add_argument(
        "--column_spacing",
        help="Set spacing between columns of markers",
        default=0,
        type=int,
    )
    tile_group.add_argument(
        "--row_num",
        help="Set number of rows of markers",
        default=1,
        type=int,
    )
    tile_group.add_argument(
        "--row_spacing",
        help="Set number of rows of markers",
        default=0,
        type=int,
    )

    # Options for CROP layout
    crop_group = parser.add_argument_group('mode: CROP')
    crop_group.add_argument(
        "--crop_size",
        help=(
            "Set dimension in mm of the cropped image in a CROP page mode, "
            "defaults to the marker border"
        ),
        default=None,
        type=int,
    )

    # Args for modifying text
    text_group = parser.add_argument_group('TEXT')
    text_group.add_argument(
        "--no_number",
        help="Do not place marker id number on the marker",
        action="store_true",
    )
    text_group.add_argument(
        "--number_size",
        help="Set the text size of the id number on the marker",
        default=10,
        type=int,
    )
    text_group.add_argument(
        "--description_format",
        type=str,
        help=(
            "Text format for the description on the marker images. "
            "`marker_id` and `marker_type` are available for string format replacement. "
            "(default: %(default)s)"
        ),
        default="{marker_type} {marker_id}",
    )
    text_group.add_argument(
        "--font",
        help="Set the text font (default: %(default)s)",
        default="calibri.ttf",
        type=str,
    )

    # Args for modifying grey border
    border_group = parser.add_argument_group('STYLE')
    border_group.add_argument(
        "--border_width",
        help="Size of the border in pixels (default: %(default)s)",
        default=1,
        type=int,
    )
    border_group.add_argument(
        "--tick_length",
        help="Length of center tick lines in pixels (default: %(default)s)",
        default=10,
        type=int,
    )

    parser.set_defaults(func=main)



import argparse
import logging

from april_vision.marker import MarkerType

from ..utils import get_tag_family

LOGGER = logging.getLogger(__name__)


def main(args: argparse.Namespace) -> None:
    tag_data = get_tag_family(args.tag_family)
    print(tag_data)


def create_subparser(subparsers: argparse._SubParsersAction) -> None:
    """Print the details about a marker family."""
    parser = subparsers.add_parser(
        "family_details",
        description="Provide the details about a marker family",
        help="Provide the details about a marker family",
    )

    parser.add_argument(
        "tag_family",
        choices=[marker.value for marker in MarkerType],
        help="Set the marker family to display info about",
    )

    parser.set_defaults(func=main)



import typing
from typing import Any, Tuple

import numpy as np
from numpy.typing import NDArray

Mat = np.ndarray[int, np.dtype]

FILE_STORAGE_APPEND: int
FILE_STORAGE_BASE64: int
FILE_STORAGE_FORMAT_AUTO: int
FILE_STORAGE_FORMAT_JSON: int
FILE_STORAGE_FORMAT_MASK: int
FILE_STORAGE_FORMAT_XML: int
FILE_STORAGE_FORMAT_YAML: int
FILE_STORAGE_INSIDE_MAP: int
FILE_STORAGE_MEMORY: int
FILE_STORAGE_NAME_EXPECTED: int
FILE_STORAGE_READ: int
FILE_STORAGE_UNDEFINED: int
FILE_STORAGE_VALUE_EXPECTED: int
FILE_STORAGE_WRITE: int
FILE_STORAGE_WRITE_BASE64: int

FONT_HERSHEY_COMPLEX: int
FONT_HERSHEY_COMPLEX_SMALL: int
FONT_HERSHEY_DUPLEX: int
FONT_HERSHEY_PLAIN: int
FONT_HERSHEY_SCRIPT_COMPLEX: int
FONT_HERSHEY_SCRIPT_SIMPLEX: int
FONT_HERSHEY_SIMPLEX: int
FONT_HERSHEY_TRIPLEX: int
FONT_ITALIC: int

CAP_PROP_APERTURE: int
CAP_PROP_ARAVIS_AUTOTRIGGER: int
CAP_PROP_AUTOFOCUS: int
CAP_PROP_AUTO_EXPOSURE: int
CAP_PROP_AUTO_WB: int
CAP_PROP_BACKEND: int
CAP_PROP_BACKLIGHT: int
CAP_PROP_BITRATE: int
CAP_PROP_BRIGHTNESS: int
CAP_PROP_BUFFERSIZE: int
CAP_PROP_CHANNEL: int
CAP_PROP_CODEC_PIXEL_FORMAT: int
CAP_PROP_CONTRAST: int
CAP_PROP_CONVERT_RGB: int
CAP_PROP_DC1394_MAX: int
CAP_PROP_DC1394_MODE_AUTO: int
CAP_PROP_DC1394_MODE_MANUAL: int
CAP_PROP_DC1394_MODE_ONE_PUSH_AUTO: int
CAP_PROP_DC1394_OFF: int
CAP_PROP_EXPOSURE: int
CAP_PROP_EXPOSUREPROGRAM: int
CAP_PROP_FOCUS: int
CAP_PROP_FORMAT: int
CAP_PROP_FOURCC: int
CAP_PROP_FPS: int
CAP_PROP_FRAME_COUNT: int
CAP_PROP_FRAME_HEIGHT: int
CAP_PROP_FRAME_WIDTH: int
CAP_PROP_GAIN: int
CAP_PROP_GAMMA: int
CAP_PROP_GIGA_FRAME_HEIGH_MAX: int
CAP_PROP_GIGA_FRAME_OFFSET_X: int
CAP_PROP_GIGA_FRAME_OFFSET_Y: int
CAP_PROP_GIGA_FRAME_SENS_HEIGH: int
CAP_PROP_GIGA_FRAME_SENS_WIDTH: int
CAP_PROP_GIGA_FRAME_WIDTH_MAX: int
CAP_PROP_GPHOTO2_COLLECT_MSGS: int
CAP_PROP_GPHOTO2_FLUSH_MSGS: int
CAP_PROP_GPHOTO2_PREVIEW: int
CAP_PROP_GPHOTO2_RELOAD_CONFIG: int
CAP_PROP_GPHOTO2_RELOAD_ON_CHANGE: int
CAP_PROP_GPHOTO2_WIDGET_ENUMERATE: int
CAP_PROP_GSTREAMER_QUEUE_LENGTH: int
CAP_PROP_GUID: int
CAP_PROP_HUE: int
CAP_PROP_IMAGES_BASE: int
CAP_PROP_IMAGES_LAST: int
CAP_PROP_INTELPERC_DEPTH_CONFIDENCE_THRESHOLD: int
CAP_PROP_INTELPERC_DEPTH_FOCAL_LENGTH_HORZ: int
CAP_PROP_INTELPERC_DEPTH_FOCAL_LENGTH_VERT: int
CAP_PROP_INTELPERC_DEPTH_LOW_CONFIDENCE_VALUE: int
CAP_PROP_INTELPERC_DEPTH_SATURATION_VALUE: int
CAP_PROP_INTELPERC_PROFILE_COUNT: int
CAP_PROP_INTELPERC_PROFILE_IDX: int
CAP_PROP_IOS_DEVICE_EXPOSURE: int
CAP_PROP_IOS_DEVICE_FLASH: int
CAP_PROP_IOS_DEVICE_FOCUS: int
CAP_PROP_IOS_DEVICE_TORCH: int
CAP_PROP_IOS_DEVICE_WHITEBALANCE: int
CAP_PROP_IRIS: int
CAP_PROP_ISO_SPEED: int
CAP_PROP_MODE: int
CAP_PROP_MONOCHROME: int
CAP_PROP_OPENNI2_MIRROR: int
CAP_PROP_OPENNI2_SYNC: int
CAP_PROP_OPENNI_APPROX_FRAME_SYNC: int
CAP_PROP_OPENNI_BASELINE: int
CAP_PROP_OPENNI_CIRCLE_BUFFER: int
CAP_PROP_OPENNI_FOCAL_LENGTH: int
CAP_PROP_OPENNI_FRAME_MAX_DEPTH: int
CAP_PROP_OPENNI_GENERATOR_PRESENT: int
CAP_PROP_OPENNI_MAX_BUFFER_SIZE: int
CAP_PROP_OPENNI_MAX_TIME_DURATION: int
CAP_PROP_OPENNI_OUTPUT_MODE: int
CAP_PROP_OPENNI_REGISTRATION: int
CAP_PROP_OPENNI_REGISTRATION_ON: int
CAP_PROP_PAN: int
CAP_PROP_POS_AVI_RATIO: int
CAP_PROP_POS_FRAMES: int
CAP_PROP_POS_MSEC: int
CAP_PROP_PVAPI_BINNINGX: int
CAP_PROP_PVAPI_BINNINGY: int
CAP_PROP_PVAPI_DECIMATIONHORIZONTAL: int
CAP_PROP_PVAPI_DECIMATIONVERTICAL: int
CAP_PROP_PVAPI_FRAMESTARTTRIGGERMODE: int
CAP_PROP_PVAPI_MULTICASTIP: int
CAP_PROP_PVAPI_PIXELFORMAT: int
CAP_PROP_RECTIFICATION: int
CAP_PROP_ROLL: int
CAP_PROP_SAR_DEN: int
CAP_PROP_SAR_NUM: int
CAP_PROP_SATURATION: int
CAP_PROP_SETTINGS: int
CAP_PROP_SHARPNESS: int
CAP_PROP_SPEED: int
CAP_PROP_TEMPERATURE: int
CAP_PROP_TILT: int
CAP_PROP_TRIGGER: int
CAP_PROP_TRIGGER_DELAY: int
CAP_PROP_VIEWFINDER: int
CAP_PROP_WB_TEMPERATURE: int
CAP_PROP_WHITE_BALANCE_BLUE_U: int
CAP_PROP_WHITE_BALANCE_RED_V: int
CAP_PROP_ZOOM: int

COLOR_BGR2GRAY: int

IMREAD_ANYCOLOR: int
IMREAD_ANYDEPTH: int
IMREAD_COLOR: int
IMREAD_GRAYSCALE: int
IMREAD_IGNORE_ORIENTATION: int
IMREAD_LOAD_GDAL: int
IMREAD_REDUCED_COLOR_2: int
IMREAD_REDUCED_COLOR_4: int
IMREAD_REDUCED_COLOR_8: int
IMREAD_REDUCED_GRAYSCALE_2: int
IMREAD_REDUCED_GRAYSCALE_4: int
IMREAD_REDUCED_GRAYSCALE_8: int
IMREAD_UNCHANGED: int

FILE_NODE_NONE: int
FILE_NODE_INT: int
FILE_NODE_REAL: int
FILE_NODE_FLOAT: int
FILE_NODE_STR: int
FILE_NODE_STRING: int
FILE_NODE_SEQ: int
FILE_NODE_MAP: int
FILE_NODE_TYPE_MASK: int
FILE_NODE_FLOW: int
FILE_NODE_UNIFORM: int
FILE_NODE_EMPTY: int
FILE_NODE_NAMED: int


def cvtColor(src: Mat, code: int, dts: Mat = ..., dstCn: int = ...) -> Mat: ...
def imread(filename: str, flags: int = ...) -> Mat: ...
def imwrite(filename: str, img: Mat, params: typing.List[int] = ...) -> bool: ...
def imencode(ext: str, img: Mat, params: typing.List[int] = ...) -> typing.Tuple[bool, Mat]: ...
def polylines(img: Mat, pts: typing.List[NDArray], isClosed: bool, color: typing.Tuple[int, int, int], thickness: int = ..., lineType: int = ..., shift: int = ...) -> Mat: ...
def putText(img: Mat, text: str, org: NDArray, fontFace: int, fontScale: float, color: typing.Tuple[int, int, int], thickness: int = ..., lineType: int = ..., bottomLeftOrigin: bool = ...) -> Mat: ...
def VideoWriter_fourcc(c1: str, c2: str, c3: str, c4: str) -> int: ...
def imshow(winname: str, mat: Mat) -> None: ...
def waitKey(delay: int = 0) -> int: ...
def destroyAllWindows() -> None: ...
def flip(src: Mat, flipCode: int) -> Mat: ...
def resize(src: Mat, dsize: Tuple[int, int]) -> Mat: ...


class Node:
    def __init__(self, payload: Any) -> None: ...
    def isSeq(self) -> bool: ...
    def isString(self) -> bool: ...
    def at(self, i0: int) -> 'Node': ...
    def size(self) -> int: ...
    def string(self) -> str: ...
    def mat(self) -> Mat: ...
    def real(self) -> float: ...


class FileStorage:
    def __init__(self, filename: str, flags: int, encoding: str = ...) -> None: ...
    def getNode(self, nodename: str) -> Node: ...
    def write(self, nodename: str, payload: typing.Any) -> None: ...
    def release(self) -> None: ...
    def startWriteStruct(self, nodename: str, flags: int) -> None: ...
    def endWriteStruct(self) -> None: ...


class VideoCapture:
    def __init__(self, index: typing.Union[str, int], apiPreference: int = ...) -> None: ...
    def set(self, propId: int, value: float) -> None: ...
    def get(self, propId: int) -> float: ...
    def release(self) -> None: ...
    def isOpened(self) -> bool: ...
    def read(self) -> typing.Tuple[bool, Mat]: ...


class VideoWriter:
    def __init__(self, filename: str, fourcc: int, fps: float, frameSize: Tuple[int, int]) -> None: ...
    def write(self, frame: Mat) -> None: ...
    def release(self) -> None: ...



"""Utilities to assist with marker operations."""
from typing import Dict, Iterable


def generate_marker_size_mapping(
    marker_sizes: Dict[Iterable[int], int],
) -> Dict[int, float]:
    """
    Unroll a dict of iterables into a dict of floats.
    To be used in pose estimation.
    """
    tag_sizes: Dict[int, float] = {}

    for marker_ids, marker_size in marker_sizes.items():
        # Unroll generators to give direct lookup
        for marker_id in marker_ids:
            # Convert to meters
            tag_sizes[marker_id] = float(marker_size) / 1000

    return tag_sizes



import logging
from pathlib import Path
from typing import Callable, Dict, Iterable, List, Optional, Union

from numpy.typing import NDArray

from april_vision import (CalibratedCamera, Frame, Marker, Processor,
                          USBCamera, __version__, calibrations, find_cameras,
                          generate_marker_size_mapping)
from april_vision.helpers import Base64Sender

LOGGER = logging.getLogger(__name__)


class AprilCamera:
    """
    Virtual Camera Board for detecting fiducial markers.

    Additionally, it will do pose estimation, along with some calibration
    in order to determine the spatial positon and orientation of the markers
    that it has detected.
    """

    name: str = "AprilTag Camera Board"

    @classmethod
    def discover(cls) -> Dict[str, 'AprilCamera']:
        """Discover boards that this backend can control."""
        return {
            (serial := f"{camera_data.name} - {camera_data.index}"):
            cls(camera_data.index, camera_data=camera_data, serial_num=serial)
            for camera_data in find_cameras(calibrations)
        }

    def __init__(self, camera_id: int, camera_data: CalibratedCamera, serial_num: str) -> None:
        """Generate a backend from the camera index and calibration data."""
        camera_source = USBCamera.from_calibration_file(
            camera_id,
            calibration_file=camera_data.calibration,
            vidpid=camera_data.vidpid,
        )
        self._cam = Processor(
            camera_source,
            calibration=camera_source.calibration,
            name=camera_data.name,
            vidpid=camera_data.vidpid,
            mask_unknown_size_tags=True,
        )
        self._serial = serial_num

    @property
    def serial_number(self) -> str:
        """Get the serial number."""
        return self._serial

    @property
    def firmware_version(self) -> Optional[str]:
        """Get the firmware version of the board."""
        return f"April camera v{__version__}"

    def make_safe(self) -> None:
        """
        Close the camera.

        The camera will no longer work after this method is called.
        """
        self._cam.close()

    # Proxy methods from USBCamera object
    def see(self, *, eager: bool = True, frame: Optional[NDArray] = None) -> List[Marker]:
        """
        Capture an image and identify fiducial markers.

        :param eager: Process the pose estimations of markers immediately,
            currently unused.
        :returns: list of markers that the camera could see.
        """
        return self._cam.see(frame=frame)

    def capture(self) -> NDArray:
        """
        Get the raw image data from the camera.

        :returns: Camera pixel data
        """
        return self._cam.capture()

    def save(self, path: Union[Path, str], *, frame: Optional[NDArray] = None) -> None:
        """Save an annotated image to a path."""
        self._cam.save(path, frame=frame)

    def set_marker_sizes(
        self,
        tag_sizes: Union[float, Dict[int, float]],
    ) -> None:
        """
        Set the size of tags that are used for pose estimation.

        If a dict is given for tag_sizes, only marker IDs that are keys of the
        dict will be detected.
        """
        self._cam.set_marker_sizes(tag_sizes)

    def set_detection_hook(self, callback: Callable[[Frame, List[Marker]], None]) -> None:
        """
        Setup a callback to be run after each dectection.
        """
        self._cam.detection_hook = callback


def setup_cameras(
    tag_sizes: Dict[Iterable[int], int],
    publish_func: Optional[Callable[[str, bytes], None]] = None,
) -> Dict[str, AprilCamera]:
    """
    Find all connected cameras with calibration and configure tag sizes.

    Optionally set a callback to send a base64 encode JPEG bytestream of each
    image detection is run on.
    """
    expanded_tag_sizes = generate_marker_size_mapping(tag_sizes)

    if publish_func:
        frame_sender = Base64Sender(publish_func)

    cameras = AprilCamera.discover()

    for camera in cameras.values():
        camera.set_marker_sizes(expanded_tag_sizes)
        if publish_func:
            camera.set_detection_hook(frame_sender.annotated_frame_hook)

    return cameras



"""
Annotate a video file.

Takes an input video, adds annotation to each frame and saves the video.
"""
import argparse
import logging
from pathlib import Path

import cv2

from ..frame_sources import VideoSource
from ..marker import MarkerType
from ..utils import annotate_text, load_calibration, normalise_marker_text
from ..vision import Processor

LOGGER = logging.getLogger(__name__)


def main(args: argparse.Namespace) -> None:
    """Annotate an video file using the provided args."""
    input_file = Path(args.input_file)
    if not input_file.exists():
        LOGGER.fatal("Input file does not exist.")
        return

    source = VideoSource(args.input_file)
    num_frames = source._video.get(cv2.CAP_PROP_FRAME_COUNT)
    LOGGER.info(f"Processing video with {num_frames:.0f} frames.")

    fps = source._video.get(cv2.CAP_PROP_FPS)
    width = int(source._video.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(source._video.get(cv2.CAP_PROP_FRAME_HEIGHT))
    calibration = None
    if args.calibration:
        _, calibration = load_calibration(args.calibration)

    processer = Processor(
        source,
        tag_family=args.tag_family,
        quad_decimate=args.quad_decimate,
        tag_sizes=float(args.tag_size) / 1000,
        calibration=calibration,
    )
    output = cv2.VideoWriter(
        args.output_file, cv2.VideoWriter_fourcc(*'mp4v'), fps, (width, height))
    while True:
        try:
            frame = processer._capture()
        except IOError:
            output.release()
            processer.close()
            break

        markers = processer._detect(frame)
        frame = processer._annotate(frame, markers)

        if args.calibration:
            try:
                for marker in markers:
                    # Check we have pose data
                    _ = marker.cartesian

                    text_scale = normalise_marker_text(marker)

                    loc = (
                        int(marker.pixel_centre.x - 80 * text_scale),
                        int(marker.pixel_centre.y + 40 * text_scale),
                    )
                    frame = annotate_text(
                        frame, f"dist={marker.distance}mm", loc,
                        text_scale=0.8 * text_scale,
                        text_colour=(255, 191, 0),  # deep sky blue
                    )
            except RuntimeError:
                pass

        # save frame
        output.write(frame.colour_frame)

        if args.preview:
            cv2.imshow('image', frame.colour_frame)
            _ = cv2.waitKey(1)

    LOGGER.info("Finished processing video.")
    if args.preview:
        cv2.destroyAllWindows()


def create_subparser(subparsers: argparse._SubParsersAction) -> None:
    """Annotate_video command parser."""
    parser = subparsers.add_parser(
        "annotate_video",
        description="Annotate a video file with its markers.",
        help="Annotate a video file with its markers.",
    )

    parser.add_argument("input_file", type=str, help="The video file to process.")
    parser.add_argument("output_file", type=str, help="The filepath to save the output to.")

    parser.add_argument(
        '--preview', action='store_true',
        help="Display the annotated video as it is annotated.")

    parser.add_argument(
        '--tag_family', default=MarkerType.APRILTAG_36H11.value,
        choices=[marker.value for marker in MarkerType],
        help="Set the marker family to detect, defaults to 'tag36h11'")
    parser.add_argument(
        '--quad_decimate', type=float, default=2,
        help="Set the level of decimation used in the detection stage")

    parser.add_argument(
        '--calibration', type=Path, default=None, help="Calbration XML file to use.")
    parser.add_argument(
        '--tag_size', type=int, default=0, help="The size of markers in millimeters")

    parser.set_defaults(func=main)



"""
Classes for marker detections and various axis representations.
"""
from enum import Enum
from math import acos, atan2, cos, degrees, sin
from typing import NamedTuple, Optional, Tuple, cast

import numpy as np
from numpy.linalg import norm as hypotenuse
from numpy.typing import NDArray
from pyapriltags import Detection
from pyquaternion import Quaternion

ThreeTuple = Tuple[float, float, float]
RotationMatrix = Tuple[ThreeTuple, ThreeTuple, ThreeTuple]


class MarkerType(Enum):
    """
    The available tag families.

    To support Apriltag 2 libraries use tag36h11.
    """

    APRILTAG_16H5 = 'tag16h5'
    APRILTAG_25H9 = 'tag25h9'
    APRILTAG_36H11 = 'tag36h11'
    APRILTAG_21H7 = 'tagCircle21h7'
    APRILTAG_49H12 = 'tagCircle49h12'
    APRILTAG_48H12 = 'tagCustom48h12'
    APRILTAG_41H12 = 'tagStandard41h12'
    APRILTAG_52H13 = 'tagStandard52h13'


class PixelCoordinates(NamedTuple):
    """
    Coordinates within an image made up from pixels.

    Floating point type is used to allow for subpixel detected locations
    to be represented.

    :param float x: X coordinate
    :param float y: Y coordinate
    """

    x: float
    y: float


class CartesianCoordinates(NamedTuple):
    """
    A 3 dimensional cartesian coordinate in the standard right-handed cartesian system.

    Origin is at the camera.

    The X axis extends directly away from the camera. Zero is at the camera.
    Increasing values indicate greater distance from the camera.

    The Y axis is horizontal relative to the camera's perspective, i.e: right
    to left within the frame of the image. Zero is at the centre of the image.
    Increasing values indicate greater distance to the left.

    The Z axis is vertical relative to the camera's perspective, i.e: down to
    up within the frame of the image. Zero is at the centre of the image.
    Increasing values indicate greater distance above the centre of the image.

    More information: https://w.wiki/5zbE

    :param float x: X coordinate, in millimeters
    :param float y: Y coordinate, in millimeters
    :param float z: Z coordinate, in millimeters
    """

    x: float
    y: float
    z: float

    @classmethod
    def from_tvec(cls, x: float, y: float, z: float) -> 'CartesianCoordinates':
        """
        Convert coordinate system to standard right-handed cartesian system.

        The pose estimation coordinate system has the origin at the camera center.
        Also converts units to millimeters.

        :param float x: The x-axis is to the right in the image taken by the camera.
        :param float y: The y-axis is down in the image taken by the camera.
        :param float z: The z-axis points from the camera center out the camera lens.
        """
        return cls(x=z * 1000, y=-x * 1000, z=-y * 1000)


class SphericalCoordinate(NamedTuple):
    """
    A 3 dimensional spherical coordinate location.

    # The conventional spherical coordinate in mathematical notation where θ is
    a rotation around the vertical axis and φ is measured as the angle from
    the vertical axis.
    More information: https://mathworld.wolfram.com/SphericalCoordinates.html

    :param float r: Radial distance from the origin, in millimeters.
    :param float theta: Azimuth angle, θ, in radians. This is the angle from
        directly in front of the camera to the vector which points to the
        location in the horizontal plane. A positive value indicates a
        counter-clockwise rotation. Zero is at the centre of the image.
    :param float phi: Polar angle, φ, in radians. This is the angle "down"
        from the vertical axis to the vector which points to the location.
        Zero is directly upward.
    """

    r: int
    theta: float
    phi: float

    @classmethod
    def from_tvec(cls, x: float, y: float, z: float) -> 'SphericalCoordinate':
        """
        Convert coordinate system to standard right-handed cartesian system.

        The pose estimation coordinate system has the origin at the camera center.

        :param float x: The x-axis is to the right in the image taken by the camera.
        :param float y: The y-axis is down in the image taken by the camera.
        :param float z: The z-axis points from the camera center out the camera lens.
        """
        _x, _y, _z = z, -x, -y

        r = hypotenuse([_x, _y, _z])
        theta = atan2(_y, _x)
        phi = acos(_z / r)
        return cls(int(r * 1000), theta, phi)


class Orientation(NamedTuple):
    """
    The orientation of an object in 3-D space.

    :param float yaw:   Get yaw of the marker, a rotation about the vertical axis, in radians.
                        Positive values indicate a rotation clockwise from the perspective of
                        the marker.
                        Zero values have the marker facing the camera square-on.
    :param float pitch: Get pitch of the marker, a rotation about the transverse axis, in
                        radians.
                        Positive values indicate a rotation upwards from the perspective of
                        the marker.
                        Zero values have the marker facing the camera square-on.
    :param float roll:  Get roll of the marker, a rotation about the longitudinal axis,
                        in radians.
                        Positive values indicate a rotation clockwise from the perspective of
                        the marker.
                        Zero values have the marker facing the camera square-on.
    """

    yaw: float
    pitch: float
    roll: float

    @classmethod
    def from_rvec_matrix(
        cls,
        rotation_matrix: NDArray,
        aruco_orientation: bool = True
    ) -> 'Orientation':
        """
        Calculate yaw, pitch, roll given the rotation matrix in the camera's coordinate system.

        More information:
        https://en.wikipedia.org/wiki/Rotation_matrix#In_three_dimensions
        """
        # Calculate the quaternion of the rotation in the camera's coordinate system
        initial_rot = Quaternion(matrix=rotation_matrix)

        return cls.from_quaternion(initial_rot, aruco_orientation)

    @classmethod
    def from_quaternion(
        cls,
        quaternion: Quaternion,
        aruco_orientation: bool = True
    ) -> 'Orientation':
        """
        Calculate yaw, pitch, roll given the quaternion in the camera's coordinate system.

        More information:
        https://en.wikipedia.org/wiki/Rotation_matrix#In_three_dimensions
        """
        # remap quaternion to global coordinate system from the token's perspective
        quaternion = Quaternion(
            quaternion.w, -quaternion.z, -quaternion.x, quaternion.y,
        )
        if aruco_orientation:
            # Rotate the quaternion so 0 roll is a marker the correct way up
            marker_orientation_correction = Quaternion(matrix=np.array([
                [1, 0, 0],
                [0, -1, 0],
                [0, 0, -1],
            ]))

            quaternion *= marker_orientation_correction

        obj = cls(*quaternion.yaw_pitch_roll)

        return obj

    @property
    def rotation_matrix(self) -> RotationMatrix:
        """
        Get the rotation matrix represented by this orientation.

        Conversion calculation: https://w.wiki/6gbp

        Returns:
            A 3x3 rotation matrix as a tuple of tuples.
        """
        psi, theta, phi = self.yaw, self.pitch, self.roll

        A12 = -cos(theta) * sin(psi) + sin(phi) * sin(theta) * cos(psi)
        A13 = sin(phi) * sin(psi) + cos(phi) * sin(theta) * cos(psi)
        A22 = cos(theta) * cos(psi) + sin(phi) * sin(theta) * sin(psi)
        A23 = -sin(phi) * cos(psi) + cos(phi) * sin(theta) * sin(psi)

        matrix = (
            (cos(theta) * cos(psi), A12, A13),
            (cos(theta) * sin(psi), A22, A23),
            (-sin(theta), sin(phi) * cos(theta), cos(phi) * cos(theta)),
        )

        return cast(RotationMatrix, matrix)

    @property
    def quaternion(self) -> Tuple[float, float, float, float]:
        """
        Get the quaternion represented by this orientation.

        Conversion calculation: https://w.wiki/6gbq

        Returns:
            A 4-tuple hamiltonian quaternion.
        """
        psi_2, theta_2, phi_2 = self.yaw / 2, self.pitch / 2, self.roll / 2
        w = cos(phi_2) * cos(theta_2) * cos(psi_2) + sin(phi_2) * sin(theta_2) * sin(psi_2)
        i = sin(phi_2) * cos(theta_2) * cos(psi_2) - cos(phi_2) * sin(theta_2) * sin(psi_2)
        j = cos(phi_2) * sin(theta_2) * cos(psi_2) + sin(phi_2) * cos(theta_2) * sin(psi_2)
        k = cos(phi_2) * cos(theta_2) * sin(psi_2) - sin(phi_2) * sin(theta_2) * cos(psi_2)

        return (w, i, j, k)


PixelCorners = Tuple[PixelCoordinates, PixelCoordinates, PixelCoordinates, PixelCoordinates]


class Marker(NamedTuple):
    """
    Wrapper of a marker detection with axis and rotation calculated.
    """

    rvec: Optional[NDArray]
    tvec: Optional[NDArray]

    id: int
    size: int
    marker_type: MarkerType
    pixel_corners: PixelCorners
    pixel_centre: PixelCoordinates

    distance: float = 0
    # In degrees, increasing clockwise
    bearing: float = 0

    cartesian: CartesianCoordinates = CartesianCoordinates(0, 0, 0)
    spherical: SphericalCoordinate = SphericalCoordinate(0, 0, 0)
    orientation: Orientation = Orientation(0, 0, 0)

    aruco_orientation: bool = True

    @classmethod
    def from_detection(
        cls,
        marker: Detection,
        *,
        aruco_orientation: bool = True,
    ) -> 'Marker':
        _tag_size = int((marker.tag_size or 0) * 1000)

        _pixel_corners = tuple(
            PixelCoordinates(x, y)
            for x, y in marker.corners.tolist()
        )
        _pixel_centre = PixelCoordinates(*marker.center.tolist())

        if marker.pose_t is not None and marker.pose_R is not None:
            _distance = int(hypotenuse(marker.pose_t) * 1000)

            _cartesian = CartesianCoordinates.from_tvec(
                *marker.pose_t.flatten().tolist()
            )
            _bearing = degrees(atan2(-_cartesian.y, _cartesian.x))

            _spherical = SphericalCoordinate.from_tvec(
                *marker.pose_t.flatten().tolist()
            )

            _orientation = Orientation.from_rvec_matrix(
                marker.pose_R,
                aruco_orientation=aruco_orientation,
            )

            return cls(
                rvec=marker.pose_R,
                tvec=marker.pose_t,
                id=marker.tag_id,
                size=_tag_size,
                marker_type=MarkerType(marker.tag_family.decode('utf-8')),
                pixel_corners=cast(PixelCorners, _pixel_corners),
                pixel_centre=_pixel_centre,
                distance=_distance,
                bearing=_bearing,
                cartesian=_cartesian,
                spherical=_spherical,
                orientation=_orientation,
                aruco_orientation=aruco_orientation,
            )
        else:
            return cls(
                rvec=None,
                tvec=None,
                id=marker.tag_id,
                size=_tag_size,
                marker_type=MarkerType(marker.tag_family.decode('utf-8')),
                pixel_corners=cast(PixelCorners, _pixel_corners),
                pixel_centre=_pixel_centre,
                aruco_orientation=aruco_orientation,
            )

    def has_pose(self) -> bool:
        return (self.rvec is not None and self.tvec is not None)

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} id={self.id} distance={self.distance:.0f}mm "
            f"bearing={self.bearing:.0f}° size={self.size}mm type={self.marker_type.name}>"
        )
